# SPDX-FileCopyrightText: 2017-2022 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

# ./blender.bin --background --python tests/python/bl_pyapi_idprop.py -- --verbose
import bpy
import rna_prop_ui
import idprop

import io
import sys
import unittest
from array import array

# Run if `numpy` is installed.
try:
    import numpy as np
except ImportError:
    np = None


class TestHelper:

    @property
    def id(self):
        return self._id

    def setUp(self):
        self._id = bpy.context.scene
        self._id.pop("cycles", None)
        assert len(self._id.keys()) == 0

    def tearDown(self):
        for key in list(self._id.keys()):
            del self._id[key]

    def assertAlmostEqualSeq(self, list1, list2):
        self.assertEqual(len(list1), len(list2))
        for v1, v2 in zip(list1, list2):
            self.assertAlmostEqual(v1, v2, places=5)


class TestIdPropertyCreation(TestHelper, unittest.TestCase):
    # Default testing idprop key identifier.
    key_id = "a"

    def test_name_empty(self):
        key_id = ""
        value = 4
        self.id[key_id] = value
        self.assertEqual(self.id[key_id], value)

    def test_name_too_long(self):
        with self.assertRaises(KeyError):
            self.id["name" * 30] = 4

    def test_int(self):
        value = 2
        self.id[self.key_id] = value
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], int))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

        with self.assertRaises(OverflowError):
            self.id[self.key_id] = 2 ** 31  # integer <= 2 ** 31-1

    def test_double(self):
        value = 2.5
        self.id[self.key_id] = value
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], float))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_unicode(self):
        value = "Hello World"
        self.id[self.key_id] = value
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], str))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_bytes(self):
        value = b"Hello World"
        self.id[self.key_id] = value
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], bytes))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_enum(self):
        # Note: Fake enum, this merely creates a PROP_INT idprop.
        value = 5
        self.id[self.key_id] = value
        self.assertEqual(self.id[self.key_id], value)
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_double_list(self):
        mylist = [1.2, 3.4, 5.6]
        self.id[self.key_id] = mylist
        self.assertEqual(self.id[self.key_id].to_list(), mylist)
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_int_list(self):
        mylist = [1, 2, 3]
        self.id[self.key_id] = mylist
        self.assertEqual(self.id[self.key_id].to_list(), mylist)
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_float_array(self):
        mylist = [1.2, 3.4, 5.6]
        self.id[self.key_id] = array("f", mylist)
        self.assertAlmostEqualSeq(self.id[self.key_id].to_list(), mylist)
        self.assertEqual(self.id[self.key_id].typecode, "f")
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_double_array(self):
        mylist = [1.2, 3.4, 5.6]
        self.id[self.key_id] = array("d", mylist)
        self.assertAlmostEqualSeq(self.id[self.key_id].to_list(), mylist)
        self.assertEqual(self.id[self.key_id].typecode, "d")
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_int_array(self):
        mylist = [1, 2, 3]
        self.id[self.key_id] = array("i", mylist)
        self.assertAlmostEqualSeq(self.id[self.key_id].to_list(), mylist)
        self.assertEqual(self.id[self.key_id].typecode, "i")
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_other_array(self):
        mylist = [1, 2, 3]
        self.id[self.key_id] = array("Q", mylist)
        self.assertEqual(self.id[self.key_id].to_list(), mylist)
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_mixed_numerical_type(self):
        self.id[self.key_id] = [1, 2, 3.4, 5]
        self.assertAlmostEqualSeq(self.id[self.key_id].to_list(), [1.0, 2.0, 3.4, 5.0])
        self.assertEqual(self.id[self.key_id].typecode, "d")
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_str_list(self):
        # I'm a bit surprised that this works
        mylist = ["abc", "qwe"]
        self.id[self.key_id] = mylist
        self.assertEqual(self.id[self.key_id], mylist)
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_sequence_mixed_type(self):
        with self.assertRaises(TypeError):
            mylist = ["abc", 3, "qwe", 3.4]
            self.id[self.key_id] = mylist

    def test_mapping_simple(self):
        mydict = {"1": 10, "2": "20", "3": 30.5}
        self.id[self.key_id] = mydict
        self.assertEqual(self.id[self.key_id]["1"], mydict["1"])
        self.assertEqual(self.id[self.key_id]["2"], mydict["2"])
        self.assertEqual(self.id[self.key_id]["3"], mydict["3"])
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_mapping_complex(self):
        mydict = {
            "1": [1, 2, 3],
            "2": {"1": "abc", "2": array("i", [4, 5, 6])},
            "3": {"1": {"1": 10}, "2": b"qwe"},
        }
        self.id[self.key_id] = mydict
        self.assertEqual(self.id[self.key_id]["1"].to_list(), [1, 2, 3])
        self.assertEqual(self.id[self.key_id]["2"]["1"], "abc")
        self.assertEqual(self.id[self.key_id]["2"]["2"].to_list(), [4, 5, 6])
        self.assertEqual(self.id[self.key_id]["3"]["1"]["1"], 10)
        self.assertEqual(self.id[self.key_id]["3"]["2"], b"qwe")
        # Complex types currently return different object types.
        # self.assertEqual(self.id[self.key_id],
        #                  self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

        with self.assertRaises(KeyError):
            a = self.id[self.key_id]["2"]["a"]

    def test_invalid_type(self):
        with self.assertRaises(TypeError):
            self.id[self.key_id] = self

    def test_rename(self):
        self.id["foo"] = {"a": 1}
        self.id["bar"] = {"b": 2}
        self.id["foo"].name = "foo"
        self.id["bar"].name = "bar"
        with self.assertRaises(NameError):
            self.id["foo"].name = "bar"


class TestIdPropertyUIData(TestHelper, unittest.TestCase):
    # Default testing idprop key identifier.
    key_id = "a"

    def test_int(self):
        value = 2
        rna_prop_ui.rna_idprop_ui_create(self.id, self.key_id, default=value)
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], int))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_double(self):
        value = 2.5
        rna_prop_ui.rna_idprop_ui_create(self.id, self.key_id, default=value)
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], float))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    def test_unicode(self):
        value = "Hello World"
        rna_prop_ui.rna_idprop_ui_create(self.id, self.key_id, default=value)
        self.assertEqual(self.id[self.key_id], value)
        self.assertTrue(isinstance(self.id[self.key_id], str))
        self.assertEqual(self.id[self.key_id],
                         self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))

    # NOTE: Bytes strings are not supported currently.

    def test_enum(self):
        value = 1
        items = [('A', '', ''), ('B', '', '')]
        rna_prop_ui.rna_idprop_ui_create(self.id, self.key_id, default=value, items=items)
        # 'Enum' 'type' of idprop currently returns integer value for direct subscription,
        # and string key for `path_resolve` usage. See also #122843 .
        self.assertEqual(self.id[self.key_id], value)
        if False:  # FIXME
            self.assertEqual(self.id[self.key_id],
                             self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))
        else:
            self.assertEqual(self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)), 'B')
        self.assertEqual(self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)), 'B')
        self.id[self.key_id] = 0
        self.assertEqual(self.id[self.key_id], 0)
        if False:  # FIXME
            self.assertEqual(self.id[self.key_id],
                             self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)))
        else:
            self.assertEqual(self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(self.key_id)), 'A')


# NOTE: the tests below are fairly deep checks on expected consistency in the py-defined dynamic RNA properties,
# between their python representations (as RNA properties) and their underlying (IDProperty-based) storage.
# See `bl_pyapi_prop.py` and `bl_pyapi_prop_array.py` for more basic but systematic testing
# of all `bpy.props`-defined property types.

# Check statically typed underlying IDProperties storage for dynamic RNA properties.
class TestIdPropertyDynamicRNA(TestHelper, unittest.TestCase):

    class TestDynRNAClass(bpy.types.PropertyGroup):
        string_prop: bpy.props.StringProperty()
        int_prop: bpy.props.IntProperty()
        float_array_prop: bpy.props.FloatVectorProperty(size=[3])
        # Basic get/set transform.
        string_prop_transform: bpy.props.StringProperty(
            default="test",
            maxlen=10,
            get_transform=lambda self, storage_val, is_set: storage_val + "!!",
            set_transform=lambda self, new_val, storage_val, is_set: storage_val + "!!" + new_val)

    def setUp(self):
        super().setUp()
        bpy.utils.register_class(self.TestDynRNAClass)
        assert type(self.id) == bpy.types.Scene
        bpy.types.Scene.dynrna_prop = bpy.props.PointerProperty(type=self.TestDynRNAClass)
        assert hasattr(self.id, "dynrna_prop")

    def tearDown(self):
        del bpy.types.Scene.dynrna_prop
        bpy.utils.unregister_class(self.TestDynRNAClass)
        super().tearDown()

    def test_dynamic_rna_basics(self):
        # The IDP_Group matching a registered RNA Pointer property to a registered
        # PropertyGroup type always exists.
        self.assertTrue('dynrna_prop' in self.id.bl_system_properties_get())
        # However, the underlying idprop data for each property of the PropertyGroup
        # does not exist untill set through an RNA property access.
        self.assertTrue(len(self.id.bl_system_properties_get()['dynrna_prop']) == 0)

        self.id.dynrna_prop.string_prop = "Test String"
        self.assertTrue(len(self.id.bl_system_properties_get()['dynrna_prop']) == 1)
        self.assertEqual(
            self.id.dynrna_prop.string_prop, self.id.bl_system_properties_get()['dynrna_prop']['string_prop'])
        self.id.bl_system_properties_get()['dynrna_prop']['string_prop'] = "String test"
        self.assertEqual(
            self.id.dynrna_prop.string_prop, self.id.bl_system_properties_get()['dynrna_prop']['string_prop'])

        self.id.dynrna_prop.int_prop = 1
        self.assertTrue(len(self.id.bl_system_properties_get()['dynrna_prop']) == 2)
        self.assertEqual(self.id.dynrna_prop.int_prop, self.id.bl_system_properties_get()['dynrna_prop']['int_prop'])
        self.id.bl_system_properties_get()['dynrna_prop']['int_prop'] = 33
        self.assertEqual(self.id.dynrna_prop.int_prop, self.id.bl_system_properties_get()['dynrna_prop']['int_prop'])

        self.id.dynrna_prop.float_array_prop = [1.0, 2.0, 3.0]
        self.assertTrue(len(self.id.bl_system_properties_get()['dynrna_prop']) == 3)
        self.assertEqual(list(self.id.dynrna_prop.float_array_prop),
                         list(self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop']))
        self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = [0.1, 1.0, 10.0]
        self.assertEqual(list(self.id.dynrna_prop.float_array_prop),
                         list(self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop']))

    def test_static_type(self):
        # Check that the idproperty behind the defined RNA ones are statically typed, and cannot be
        # assigned random other types of values.
        self.id.dynrna_prop.string_prop = "Test String"
        self.id.dynrna_prop.int_prop = 1
        self.id.dynrna_prop.float_array_prop = [1.0, 2.0, 3.0]

        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['string_prop'] = b"Test String"

        # Supported conversion from bool to int.
        self.id.bl_system_properties_get()['dynrna_prop']['int_prop'] = False
        self.assertEqual(self.id.bl_system_properties_get()['dynrna_prop']['int_prop'], 0)
        self.assertEqual(self.id.dynrna_prop.int_prop, self.id.bl_system_properties_get()['dynrna_prop']['int_prop'])
        self.assertEqual(type(self.id.bl_system_properties_get()['dynrna_prop']['int_prop']), int)
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['int_prop'] = 2.5
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['int_prop'] = "100"

        mixed_array = [5, 2.5, 2]
        self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = mixed_array
        self.assertEqual(list(self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop']), mixed_array)
        self.assertEqual(list(self.id.dynrna_prop.float_array_prop),
                         list(self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop']))
        self.assertTrue(
            all((type(i) is float for i in self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'])))
        # Assign out-of int32 range value to a float property.
        self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = [1000000000000, 5, 6]
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = 2.5
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = [True, False, False]
        # Length of the array is also considered as static/fixed.
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = [5.0, 2.5]
        with self.assertRaises(TypeError):
            self.id.bl_system_properties_get()['dynrna_prop']['float_array_prop'] = [1.0, 10.0, 100.0, 0.1]

    def test_get_set_transform(self):
        self.assertEqual(len(self.id.dynrna_prop.string_prop_transform), 6)
        self.assertEqual(self.id.dynrna_prop.string_prop_transform, "test!!")
        # Default value only, was not yet set.
        self.assertFalse('string_prop_transform' in self.id.bl_system_properties_get()['dynrna_prop'])

        self.id.dynrna_prop.string_prop_transform = "-"
        self.assertEqual(self.id.dynrna_prop.string_prop_transform, "test!!-!!")
        self.assertEqual(self.id.bl_system_properties_get()['dynrna_prop']['string_prop_transform'], "test!!-")

        # Raw-set exactly maxlen - 1 char.
        self.id.bl_system_properties_get()['dynrna_prop']['string_prop_transform'] = "test!!tes"
        # get_transform will produce an 11-char results, which should trigger an error.
        # These asserts are not raised currently, but only print in `stderr`...
        # But the returned string is the 'storage' one, not the result from get_transform.
        stderr, sys.stderr = sys.stderr, io.StringIO()
        self.assertEqual(self.id.dynrna_prop.string_prop_transform, "test!!tes")
        self.assertTrue("ValueError" in sys.stderr.getvalue() and
                        "10" in sys.stderr.getvalue() and "11" in sys.stderr.getvalue())
        sys.stderr.close()
        sys.stderr = stderr

        # Raw-set back to default value.
        self.id.bl_system_properties_get()['dynrna_prop']['string_prop_transform'] = "test"
        # Now set_transform will return 12-char string, wich is also invalid and discarded.
        stderr, sys.stderr = sys.stderr, io.StringIO()
        self.id.dynrna_prop.string_prop_transform = "test!!"
        self.assertTrue("ValueError" in sys.stderr.getvalue() and
                        "10" in sys.stderr.getvalue() and "12" in sys.stderr.getvalue())
        sys.stderr.close()
        sys.stderr = stderr
        self.assertEqual(self.id.bl_system_properties_get()['dynrna_prop']['string_prop_transform'], "test")
        self.assertEqual(self.id.dynrna_prop.string_prop_transform, "test!!")


class TestIdPropertyGroupView(TestHelper, unittest.TestCase):

    def test_type(self):
        self.assertEqual(type(self.id.keys()), idprop.types.IDPropertyGroupViewKeys)
        self.assertEqual(type(self.id.values()), idprop.types.IDPropertyGroupViewValues)
        self.assertEqual(type(self.id.items()), idprop.types.IDPropertyGroupViewItems)

        self.assertEqual(type(iter(self.id.keys())), idprop.types.IDPropertyGroupIterKeys)
        self.assertEqual(type(iter(self.id.values())), idprop.types.IDPropertyGroupIterValues)
        self.assertEqual(type(iter(self.id.items())), idprop.types.IDPropertyGroupIterItems)

    def test_basic(self):
        text = ["A", "B", "C"]
        for i, ch in enumerate(text):
            self.id[ch] = i
        self.assertEqual(len(self.id.keys()), len(text))
        self.assertEqual(list(self.id.keys()), text)
        self.assertEqual(list(reversed(self.id.keys())), list(reversed(text)))

        self.assertEqual(len(self.id.values()), len(text))
        self.assertEqual(list(self.id.values()), list(range(len(text))))
        self.assertEqual(list(reversed(self.id.values())), list(reversed(range(len(text)))))

        self.assertEqual(len(self.id.items()), len(text))
        self.assertEqual(list(self.id.items()), [(k, v) for v, k in enumerate(text)])
        self.assertEqual(list(reversed(self.id.items())), list(reversed([(k, v) for v, k in enumerate(text)])))

        # Check direct iteration is working as expected.
        self.id["group"] = {ch: i for i, ch in enumerate(text)}
        group = self.id["group"]

        self.assertEqual(len(group), len(text))
        self.assertEqual(list(iter(group)), text)

    def test_contains(self):
        # Check `idprop.types.IDPropertyGroupView{Keys/Values/Items}.__contains__`
        text = ["A", "B", "C"]
        for i, ch in enumerate(text):
            self.id[ch] = i

        self.assertIn("A", self.id)
        self.assertNotIn("D", self.id)

        self.assertIn("A", self.id.keys())
        self.assertNotIn("D", self.id.keys())

        self.assertIn(2, self.id.values())
        self.assertNotIn(3, self.id.values())

        self.assertIn(("A", 0), self.id.items())
        self.assertNotIn(("D", 3), self.id.items())


class TestBufferProtocol(TestHelper, unittest.TestCase):

    def test_copy(self):
        self.id["a"] = array("i", [1, 2, 3, 4, 5])
        self.id["b"] = self.id["a"]
        self.assertEqual(self.id["a"].to_list(), self.id["b"].to_list())

    def test_memview_attributes(self):
        mylist = [1, 2, 3]
        self.id["a"] = mylist

        view1 = memoryview(self.id["a"])
        view2 = memoryview(array("i", mylist))

        self.assertEqualMemviews(view1, view2)

    def assertEqualMemviews(self, view1, view2):
        props_to_compare = (
            "contiguous", "format", "itemsize", "nbytes", "ndim",
            "readonly", "shape", "strides", "suboffsets",
        )
        for attr in props_to_compare:
            self.assertEqual(getattr(view1, attr), getattr(view2, attr))

        self.assertEqual(list(view1), list(view2))
        self.assertEqual(view1.tobytes(), view2.tobytes())


if np is not None:
    class TestBufferProtocol_Numpy(TestHelper, unittest.TestCase):
        def test_int(self):
            self.id["a"] = array("i", [1, 2, 3, 4, 5])
            a = np.frombuffer(self.id["a"], self.id["a"].typecode)
            self.assertEqual(len(a), 5)
            a[2] = 10
            self.assertEqual(self.id["a"].to_list(), [1, 2, 10, 4, 5])

        def test_float(self):
            self.id["a"] = array("f", [1.0, 2.0, 3.0, 4.0])
            a = np.frombuffer(self.id["a"], self.id["a"].typecode)
            self.assertEqual(len(a), 4)
            a[-1] = 10
            self.assertEqual(self.id["a"].to_list(), [1.0, 2.0, 3.0, 10.0])

        def test_double(self):
            self.id["a"] = array("d", [1.0, 2.0, 3.0, 4.0])
            a = np.frombuffer(self.id["a"], self.id["a"].typecode)
            a[1] = 10
            self.assertEqual(self.id["a"].to_list(), [1.0, 10.0, 3.0, 4.0])

        def test_full_update(self):
            self.id["a"] = array("i", [1, 2, 3, 4, 5, 6])
            a = np.frombuffer(self.id["a"], self.id["a"].typecode)
            a[:] = [10, 20, 30, 40, 50, 60]
            self.assertEqual(self.id["a"].to_list(), [10, 20, 30, 40, 50, 60])

        def test_partial_update(self):
            self.id["a"] = array("i", [1, 2, 3, 4, 5, 6, 7, 8])
            a = np.frombuffer(self.id["a"], self.id["a"].typecode)
            a[1:5] = [10, 20, 30, 40]
            self.assertEqual(self.id["a"].to_list(), [1, 10, 20, 30, 40, 6, 7, 8])


class TestRNAData(TestHelper, unittest.TestCase):

    def test_custom_properties_access(self):
        # Ensure the RNA path resolving behaves as expected & is compatible with ID-property keys.
        keys_to_test = (
            "test",
            "\\"
            '"',
            '""',
            '"""',
            '[',
            ']',
            '[]',
            '["]',
            '[""]',
            '["""]',
            '[""""]',
            # Empty properties are also valid.
            "",
        )
        for key_id in keys_to_test:
            self.id[key_id] = 1
            self.assertEqual(self.id[key_id], self.id.path_resolve('["%s"]' % bpy.utils.escape_identifier(key_id)))
            del self.id[key_id]

    def test_custom_properties_none(self):
        bpy.data.objects.new("test", None)
        test_object = bpy.data.objects["test"]

        # Access default RNA data values.
        test_object.id_properties_clear()
        test_object["test_prop"] = 0.5
        ui_data_test_prop = test_object.id_properties_ui("test_prop")

        rna_data = ui_data_test_prop.as_dict()
        self.assertTrue("min" in rna_data)
        self.assertLess(rna_data["min"], -10000.0)
        self.assertEqual(rna_data["subtype"], "NONE")
        self.assertGreater(rna_data["soft_max"], 10000.0)

        # Change RNA data values.
        ui_data_test_prop.update(subtype="TEMPERATURE", min=0, soft_min=0.1)
        rna_data = ui_data_test_prop.as_dict()
        self.assertEqual(rna_data["min"], 0)
        self.assertEqual(rna_data["soft_min"], 0.1)
        self.assertEqual(rna_data["subtype"], "TEMPERATURE")

        # Copy RNA data values from one property to another.
        test_object["test_prop_2"] = 11.7
        ui_data_test_prop_2 = test_object.id_properties_ui("test_prop_2")
        ui_data_test_prop_2.update_from(ui_data_test_prop)
        rna_data = ui_data_test_prop_2.as_dict()
        self.assertEqual(rna_data["min"], 0)
        self.assertEqual(rna_data["soft_min"], 0.1)
        self.assertEqual(rna_data["subtype"], "TEMPERATURE")
        self.assertGreater(rna_data["soft_max"], 10000.0)

        # Copy RNA data values to another object's property.
        bpy.data.objects.new("test_2", None)
        test_object_2 = bpy.data.objects["test_2"]
        test_object_2["test_prop_3"] = 20.1
        ui_data_test_prop_3 = test_object_2.id_properties_ui("test_prop_3")
        ui_data_test_prop_3.update_from(ui_data_test_prop_2)
        rna_data = ui_data_test_prop_3.as_dict()
        self.assertEqual(rna_data["min"], 0)
        self.assertEqual(rna_data["soft_min"], 0.1)
        self.assertEqual(rna_data["subtype"], "TEMPERATURE")
        self.assertGreater(rna_data["soft_max"], 10000.0)

        # Test RNA data for string property.
        test_object.id_properties_clear()
        test_object["test_string_prop"] = "Hello there!"
        ui_data_test_prop_string = test_object.id_properties_ui("test_string_prop")
        ui_data_test_prop_string.update(default="Goodbye where?")
        rna_data = ui_data_test_prop_string.as_dict()
        self.assertEqual(rna_data["default"], "Goodbye where?")

        # Test RNA data for array property.
        test_object.id_properties_clear()
        test_object["test_array_prop"] = [1, 2, 3]
        ui_data_test_prop_array = test_object.id_properties_ui("test_array_prop")
        ui_data_test_prop_array.update(default=[1, 2])
        rna_data = ui_data_test_prop_array.as_dict()
        self.assertEqual(rna_data["default"], [1, 2])

        # Test RNA data for enum property.
        test_object.id_properties_clear()
        test_object["test_enum_prop"] = 2
        ui_data_test_prop_enum = test_object.id_properties_ui("test_enum_prop")
        ui_data_test_prop_enum_items = [
            ("TOMATOES", "Tomatoes", "Solanum lycopersicum"),
            ("CUCUMBERS", "Cucumbers", "Cucumis sativus"),
            ("RADISHES", "Radishes", "Raphanus raphanistrum"),
        ]
        ui_data_test_prop_enum.update(items=ui_data_test_prop_enum_items)
        ui_data_test_prop_enum_items_full = [
            ("TOMATOES", "Tomatoes", "Solanum lycopersicum", 0, 0),
            ("CUCUMBERS", "Cucumbers", "Cucumis sativus", 0, 1),
            ("RADISHES", "Radishes", "Raphanus raphanistrum", 0, 2),
        ]
        rna_data = ui_data_test_prop_enum.as_dict()
        self.assertEqual(rna_data["items"], ui_data_test_prop_enum_items_full)


if __name__ == '__main__':
    import sys
    sys.argv = [__file__] + (sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else [])
    unittest.main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2015-2025 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import sys
import subprocess

from pathlib import Path
from modules import render_report


def get_movie_file_suffix(filepath):
    """
    Get suffix used for the video output.
    The script does not have access to the .blend file content, so deduct it from the .blend filename.
    """

    return Path(filepath).stem.split("_")[-1]


def get_arguments(filepath, output_filepath):
    suffix = get_movie_file_suffix(filepath)

    args = [
        "--background",
        "--factory-startup",
        "--enable-autoexec",
        "--debug-memory",
        "--debug-exit-on-error",
        filepath,
        "-o", f"{output_filepath}.{suffix}",
        "-a",
    ]

    return args


def create_argparse():
    parser = argparse.ArgumentParser(
        description="Run test script for each blend file in TESTDIR, comparing the render result with known output."
    )
    parser.add_argument("--blender", required=True)
    parser.add_argument("--testdir", required=True)
    parser.add_argument("--outdir", required=True)
    parser.add_argument("--oiiotool", required=True)
    parser.add_argument("--batch", default=False, action="store_true")
    return parser


class VideoOutputReport(render_report.Report):
    def postprocess_test(self, blender, test):
        suffix = get_movie_file_suffix(test.filepath)

        video_file = Path(f"{test.tmp_out_img_base}.{suffix}").as_posix()

        # If oiiotool supports the FFmpeg this could be used instead.
        """
        command = (
            self.oiiotool,
            "-i", video_file,
            "-o", test.tmp_out_img,
        )

        try:
            subprocess.check_output(command)
        except subprocess.CalledProcessError as e:
            pass
        """

        # Blender's render pipeline always appends frame suffix unless # is present in the file path.
        # Here we need the file name to match exactly, so we trick Blender by going 0001 -> #### mask
        # allowing render piepline to expand it back to 0001.
        out_filepath = test.tmp_out_img.replace("0001", "####")

        python_expr = (
            f"""
import bpy
scene = bpy.context.scene
scene.render.resolution_x = 1920
scene.render.resolution_y = 1080
scene.render.resolution_percentage = 25
ed = scene.sequence_editor_create()
strip = ed.strips.new_movie(name='input', filepath='{video_file}', channel=1, frame_start=1)
strip.colorspace_settings.name = 'Non-Color'
""")

        command = (
            blender,
            "--background",
            "--factory-startup",
            "--enable-autoexec",
            "--python-expr", python_expr,
            "-o", out_filepath,
            "-F", "PNG",
            "-f", "1",
        )

        try:
            subprocess.check_output(command)
        except subprocess.CalledProcessError as e:
            pass


def main():
    parser = create_argparse()
    args = parser.parse_args()

    report = VideoOutputReport("Sequencer", args.outdir, args.oiiotool)
    report.set_pixelated(True)
    # Default error tolerances are quite large, lower them.
    report.set_fail_threshold(2.0 / 255.0)
    report.set_fail_percent(0.01)
    report.set_reference_dir("reference")

    ok = report.run(args.testdir, args.blender, get_arguments, batch=args.batch)

    sys.exit(not ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This script generates a graphic of Blender's code layout:

./blender.bin -b --factory-startup --python ./tools/utils_doc/code_layout_diagram.py
"""
# This is an update to the historic: https://download.blender.org/ftp/ideasman42/pics/code_layout_historic.jpg
# Originally located at: `www.blender.org/bf/codelayout.jpg` (now broken).

__all__ = (
    "main",
)

from dataclasses import dataclass

import os
import sys
import bpy

from mathutils import (
    Euler,
    Vector,
)


# -----------------------------------------------------------------------------
# Generic Globals

BASE_DIR = os.path.normpath(os.path.dirname(__file__))
ROOT_DIR = os.path.normpath(os.path.join(BASE_DIR, "..", ".."))


# -----------------------------------------------------------------------------
# Data

PAGE_WIDTH = 4.5
PAGE_WIDTH_HALF = PAGE_WIDTH / 2.0

# RESOLUTION_SCALE = 0.25  # For quick test renders.
RESOLUTION_SCALE = 1.0
RESOLUTION_X = 3200

OUTPUT_IMAGE_FILE = True
OUTPUT_BLEND_FILE = False

# Can't use `Inter.woff2` as it contains overlapping glyphs.
FONT_FILE_DEFAULT = os.path.join(ROOT_DIR, "release", "datafiles", "fonts", "Noto Sans CJK Regular.woff2")
FONT_FILE_MONO = os.path.join(ROOT_DIR, "release", "datafiles", "fonts", "DejaVuSansMono.woff2")

# Apply some offsets, needed as the "default" font has strange scaling.
FONT_STYLE_SETTINGS = {
    "default": {"scale": 1.8, "space_line": 0.5},
    "mono": {"scale": 1.0, "space_line": 1.0},
}

VFONT_FROM_STYLE = {
    "default": None,
    "mono": None,
}

MATERIAL_FROM_COLOR = {
    "black": None,
    "grey": None,
}

MATERIAL_INDEX_BLACK = 0
MATERIAL_INDEX_GREY = 1

# -----------------------------------------------------------------------------
# Graphics

ARROW_DOWN = ((0, 0), (-1, 1.5), (-0.3, 1.5), (-0.3, 3.3), (0.3, 3.3), (0.3, 1.5), (1, 1.5),)
ARROW_DOWN_AND_SIDEWAYS = (
    (0, 0), (-1, 1.5), (-0.3, 1.5), (-0.3, 2.7), (-1.5, 2.7), (-1.5, 2), (-3, 3), (-1.5, 4), (-1.5, 3.3), (1.5, 3.3),
    (1.5, 4), (3, 3), (1.5, 2), (1.5, 2.7), (0.3, 2.7), (0.3, 1.5), (1, 1.5),
)


# -----------------------------------------------------------------------------
# Directory Layout Definition

@dataclass
class SectionData:
    heading: str
    source_code_base: str
    # Don't draw any arrows when None.
    source_code_call_sibling_modules: bool | None
    source_code_dirs: list


SECTIONS = (
    SectionData(
        heading="Application startup",
        source_code_call_sibling_modules=False,
        source_code_base="source/",
        source_code_dirs=[
            ("creator", "Blender's main() function, initialization & argument handling."),
            ("blender", "Contains the majority of Blender's functionality."),
        ],
    ),
    SectionData(
        heading="Editor definitions, drawing, interaction",
        source_code_call_sibling_modules=True,
        source_code_base="source/blender/editors/",
        source_code_dirs=[
            ("space_action", "Animation editor for actions."),
            ("space_buttons", "Properties editor."),
            ("space_clip", "Movie clip editor."),
            ("space_console", "Python interactive console."),
            ("space_file", "File selection."),
            ("space_graph", "Animation editor for F-Curves."),
            ("space_image", "Image editor & texture painting."),
            ("space_info", "Info space (for logging)."),
            ("space_nla", "Animation non-linear-action editor."),
            ("space_node", "Node editor for compositor, geometry and other nodes."),
            ("space_outliner", "Outliner editor."),
            ("space_script", "Unused script space (historic)."),
            ("space_sequencer", "Video sequence editor."),
            ("space_spreadsheet", "Spreadsheet data editor."),
            ("space_statusbar", "Window status bar."),
            ("space_text", "Text editor."),
            ("space_topbar", "Window file menu."),
            ("space_userpref", "User preferences."),
            ("space_view3d", "3D viewport."),
        ],
    ),
    SectionData(
        heading="Editor utilities",
        source_code_call_sibling_modules=True,
        source_code_base="source/blender/editors/",
        source_code_dirs=[
            ("datafiles", "Definitions for data-files used by editors: icons, material previews & startup file."),
            ("gizmo_library", "Shared shapes for gizmo drawing."),
            ("screen", "Screen operators & manipulation, including windowing operations."),
            ("space_api", "Space shared API's."),
            ("undo", "Undo operators and high-level API."),
            ("util", "Shared utilities including numeric input, image access, gizmo drawing & selection logic."),
        ],
    ),
    SectionData(
        heading="Editor Tools",
        source_code_call_sibling_modules=True,
        source_code_base="source/blender/editors/",
        source_code_dirs=[
            ("animation", "Utilities for key-framing and animation editors."),
            ("armature", "Tools for editing armature object data."),
            ("asset", "Tools for the asset editors."),
            ("curve", "Legacy curve object support."),
            ("curves", "Curve object support."),
            ("geometry", "Generic utilities for dealing with geometry (meshes, curves ... etc)."),
            ("gpencil_legacy", "Tools for editing grease pencil, legacy now only used for \"Annotations\"."),
            ("grease_pencil", "Operators and utilities to manipulate grease mencil."),
            ("id_management", "Operators and utilities for managing ID data blocks."),
            ("include", "Shared includes for editor modules which share logic with other editors."),
            ("interface", "Graphical user interface logic for widgets and their interactions."),
            ("io", "Operators for importing & exporting 3D content such as geometry, grease pencil & animation."),
            ("lattice", "Tools for editing lattice object data."),
            ("mask", "Tools for editing and animating 2D masks."),
            ("mesh", "Tools for editing mesh object data."),
            ("metaball", "Tools for editing meta-ball object data."),
            ("object", "Tools for editing objects and collections."),
            ("physics", "Tools for physics including particles, rigidbody & dynamic paint."),
            ("pointcloud", "Tools for point-cloud manipulation."),
            ("render", "Render operators, viewing & viewport preview."),
            ("scene", "Tools for scene manipulation (add/copy/remove)."),
            ("sculpt_paint", "Tools for sculpting and painting."),
            ("sound", "Tools for sound manipulation, packing, unpacking & mixing down audio."),
            ("transform", "Interactive transform and transform implementations (rotate, scale, move ... etc)."),
            ("uvedit", "Mesh UV editing tools for copy/paste, selection & packing."),
        ],
    ),
    SectionData(
        heading="Window, events, operators, core interaction",
        source_code_call_sibling_modules=False,
        source_code_base="source/blender/",
        source_code_dirs=[
            ("windowmanager", "General window, event handling."),
        ],
    ),
    SectionData(
        heading="General Blender API's",
        source_code_call_sibling_modules=True,
        source_code_base="source/blender/",
        source_code_dirs=[
            ("asset_system", "Asset system back-end (asset representations, asset libraries, catalogs, etc.)."),
            ("blendthumb", "Thumbnail extraction for the BLEND file-format."),
            ("blenfont", "Font loading and rendering. Used by the user-interface and stamping text into images."),
            ("animrig", "Animation & rigging functionality including key-framing, drivers NLA & actions."),
            ("blenkernel",
             "Kernel functions "
             "(data structure manipulation, allocation, free. No interactive tools or UI logic, very low level)."),
            ("blenlib",
             "Internal misc libraries: "
             "math functions, lists, random, noise, memory pools, file operations (platform agnostic)."),
            ("blenloader", "Blend file loading and writing as well as in memory undo file system."),
            ("blenloader_core", "Core blend file access (shared by Blender & blend-thumbnail extraction)."),
            ("blentranslation", "Internal support for non-English translations (gettext)."),
            ("bmesh", "Mesh-data manipulation. Used by mesh edit-mode."),
            ("compositor", "Image compositor which implements the compositor and various nodes."),
            ("cpucheck", "Support detecting if the CPU is capable of running Blender."),
            ("datatoc", "Utility used by the build-system to convert data-files into source."),
            ("depsgraph",
             "Dependency graph API. Used to track relations between various pieces of data in a Blender file."),
            ("draw", "Viewport drawing and Eevee."),
            ("editors", "Logic for (expanded on in other sections)."),
            ("freestyle", "Freestyle NPR rendering engine (scene graph, Python code, stroke, geometry.. etc)."),
            ("functions",
             "Run-time type system that implements lazy functions, multi-function network & generic functions. "
             "Used by geometry nodes."),
            ("geometry", "Low level geometry functionality. Used by object-modifiers and geometry nodes."),
            ("gpu", "Abstract graphics API's (OpenGL, Metal & Vulkan)."),
            ("ikplugin", "Generic API for inverse-kinematics (IK) solvers. Used by the IK constraint."),
            ("imbuf", "Image buffer API for image manipulation."),
            ("io", "Input/output libraries for various format (including 2D & 3D file formats)."),
            ("makesdna", "Extracting data-definitions stored in the BLEND file-format."),
            ("makesrna", "Defines access method for DNA. Used by the user-interface, Python API & animation system."),
            ("modifiers", "Object modifiers."),
            ("nodes", "Nodes code: `CMP`: composite, `GEO`: geometry, `SHD`: material, `TEX`: texture."),
            ("python", "Python API integration."),
            ("render", "Rendering & baking, integrates various rendering engines, outputs to image & video formats."),
            ("sequencer", "Video sequence editor."),
            ("shader_fx", "Grease pencil effects."),
            ("simulation", "Hair simulation."),
        ],
    ),
    SectionData(
        heading="Internal Utilities (maintained internally)",
        source_code_call_sibling_modules=False,
        source_code_base="intern/",
        source_code_dirs=[
            ("atomic", "Low level operations lockless concurrent programming."),
            ("audaspace", "Wrapper (see extern)."),
            ("clog", "C-logging library."),
            ("uriconvert", "Support encoding strings (typically paths) as URI's."),
            ("cycles", "Cycles rendering engine."),
            ("dualcon", "Re-meshing "),
            ("eigen", "Wrapper (see extern/Eigen3/)"),
            ("ghost", "Platform abstraction for windowing and user (mouse, keyboard) input."),
            ("guardedalloc", "Memory allocator used across most of Blender's internal code."),
            ("iksolver", "An IK-solver (useful for character animation)"),
            ("itasc", "An IK-solver (useful for robotics)."),
            ("libc_compat", "C-library compatibility (Linux only)."),
            ("libmv", "Movie clip motion tracking solver."),
            ("mantaflow", "Wrapper (see extern)."),
            ("memutil", "Memory allocation utilities."),
            ("mikktspace", "Calculation for mesh tangents from normals & UV's."),
            ("opensubdiv", "Wrapper (see libraries)."),
            ("openvdb", "OpenVDB wrapper for volumetric data support."),
            ("quadriflow", "Wrapper for quadriflow re-meshing."),
            ("renderdoc_dynload", "Dynamic loader (see libraries)."),
            ("rigidbody", "Wrapper for bullet physics (see extern)."),
            ("sky", "Calculate skylight & solar radiance models."),
            ("slim", "Calculate UV unwrapping, used for the \"Minimum Stretch\" method."),
            ("utfconv", "Unicode text conversion."),
            ("wayland_dynload", "Dynamic loader for wayland (Unix only)."),
        ],
    ),
    SectionData(
        heading="External Utilities (maintained externally)",
        source_code_call_sibling_modules=False,
        source_code_base="extern/",
        source_code_dirs=[
            ("audaspace", "A high level audio library for portable audio output."),
            ("binreloc", "Support access binary locations (Linux only)."),
            ("bullet2", "Physics solver with rigid-body support."),
            ("ceres", "A library for modeling and solving large, complicated optimization problems."),
            ("cuew", "Extension wrangler for CUDA."),
            ("curve_fit_nd", "Fit bezier-splines to sampled points. Used for free-hand drawing."),
            ("draco", "Compressed 3D file format support."),
            ("Eigen3", "A C++ template library for linear algebra."),
            ("fast_float", "Fast floating point number parsing."),
            ("fmtlib", "A modern C++ string formatting library."),
            ("gflags", "A library to parse command-line flags. Used by glog."),
            ("glew-es", "Extension wrangler for GLEW-ES."),
            ("glog", "Google C++ logging library. Used by gtest & ceres."),
            ("gmock", "Google C++ mocking library. Used by gtest."),
            ("gtest", "Google C++ testing framework. Used for C & C++ tests."),
            ("hipew", "Wrapper for the HIP C++ library, portability library for AMD & NVIDIA GPUs."),
            ("json", "JSON file format support."),
            ("lzma", "High quality but slower (de)compression library."),
            ("lzo", "Fast (de)compression library."),
            ("mantaflow", "Fluid simulation."),
            ("nanosvg", "Salable vector graphics support."),
            ("quadriflow", "Remeshing that rebuilds the geometry with a more uniform topology."),
            ("rangetree", "Efficient range storage."),
            ("renderdoc", "Graphical GPU debugger."),
            ("tinygltf", "GLTF 3D file format support."),
            ("ufbx", "FBX 3D file format loading support, used by the C++ FBX importer."),
            ("vulkan_memory_allocator", "Memory allocation utilities for Vulkan GPU back-end."),
            ("wcwidth", "Unicode character width."),
            ("xdnd", "Drag & drop support for the X11 graphical environment."),
            ("wintab", "Tablet input device support for MS-Windows."),
            ("xxhash", "Fast non-cryptographic hashing functions."),
        ],
    ),
    SectionData(
        heading="Pre-compiled Libraries (in SVN, or require install)",
        source_code_call_sibling_modules=False,
        source_code_base="lib/{platform}/",
        source_code_dirs=[
            ("alembic", "An interchangeable computer graphics file format for baked mesh data."),
            ("brotli", "Compression library used to decompress WOFF2 fonts."),
            ("dpcpp", "OneAPI DPC++/C++ Compiler. Used by Cycles oneAPI."),
            ("embree", "A collection of high-performance ray tracing kernels. Used by Cycles."),
            ("epoxy", "Extension wrangler for OpenGL."),
            ("ffmpeg", "Movie encoding/decoding library."),
            ("fftw3", "Fast fourier transformation library."),
            ("freetype", "Font reading & rendering library."),
            ("fribidi",
             "Bi-directional text support (right-to-left) for Arabic & Hebrew script. "
             "Intended for complex text shaping (not yet supported)."),
            ("gmp", "Arbitrary precision arithmetic library."),
            ("harfbuzz", "Text shaping engine for complex script."),
            ("haru", "PDF generation library."),
            ("hiprt", "Ray-tracing for AMD GPU's. Used by Cycles."),
            ("imath", "Library used by OpenEXR image-format."),
            ("jemalloc", "An improved memory allocator."),
            ("jpeg", "JPEG image-format support."),
            ("level-zero", "OneAPI loader & validation. Used by Cycles oneAPI."),
            ("llvm", "Low level virtual machine. Used by OSL."),
            ("manifold", "A library for operating on manifold meshes. Used as a boolean solver."),
            ("materialx", "A standard for representing materials. Used by USD, Hydra & Blender's shader nodes."),
            ("mesa", "Used for it's software OpenGL implementation."),
            ("openal", "Cross platform audio output."),
            ("opencolorio", "A solution for highly precise, performant, and consistent color management."),
            ("openexr", "EXR image-format support."),
            ("openimagedenoise", "Denoising filters for images rendered with ray tracing. Used by Cycles."),
            ("openimageio", "A library for reading, writing, and processing images in a wide variety of file formats."),
            ("openjpeg", "JPEG image-format support."),
            ("openpgl", "Intel open path guiding library. Used by Cycles."),
            ("opensubdiv", "Subdivision surface support for meshes."),
            ("openvdb", "Volumetric file-format support."),
            ("osl", "Open Shading Language. Used by Cycles."),
            ("png", "PNG image-format support."),
            ("potrace", "Trace bitmap images to vectors."),
            ("pugixml", "Light-weight C++ XML processing library."),
            ("python", "Python scripting language."),
            ("sdl", "Simple DirectMedia Layer to abstract platform specific code."),
            ("shaderc", "Shader compilation utilities. Used for GLSL shaders to be used with Vulkan."),
            ("sndfile", "Sound encoding/decoding library."),
            ("spnav", "3D mouse support, also known as NDOF. (Unix only)."),
            ("tbb", "Threading Building Blocks (C++ library)."),
            ("tiff", "TIFF image-format support."),
            ("usd", "Universal Scene Description for describing, composing, simulating, & collaborating 3D worlds."),
            ("wayland", "Wayland scanner (Unix only)."),
            ("wayland-protocols", "Wayland protocol definitions (Unix only)."),
            ("wayland_libdecor", "Wayland windowing (Unix only)."),
            ("wayland_weston", "Wayland compositor, used for running headless UI tests (Unix only)."),
            ("webp", "Support for the WEBP image format."),
            ("xml2", "XML encoding / decoding library."),
            ("xr_openxr_sdk", "Virtual reality library."),
            ("zlib", "ZLIB (GZIP) compression library."),
            ("zstd", "Z-standard compression library."),
            ("vulkan", "The Vulkan graphics API. Used by the Vulkan GPU back-end.")
        ],
    ),
    SectionData(
        heading="Operating system",
        source_code_call_sibling_modules=None,
        source_code_base="",
        source_code_dirs=[
            ("GPU Driver", "OpenGL (or Metal on Apple)."),
            ("Standard C", "Also known as libc."),
            ("C++ & STL", "C++ Standard Library."),
            ("Windowing", "Cocoa on Apple.\nX11/Wayland on Unix.\nWIN32 on MS-Window."),
        ],
    ),

    SectionData(
        heading="Other Directories",
        source_code_call_sibling_modules=None,
        source_code_base="",
        source_code_dirs=[
            ("assets", "Bundled assets such as brushes & nodes, accessible from the default installation."),
            ("build_files", "Files used by CMake, the build-bot & utilities for packaging blender builds."),
            ("doc", "Scripts for building Blender's Python's API docs, man-page and DOXYGEN documentation."),
            ("locale", "Translations for Blender's interface."),
            ("release", "Files bundled with Blender including fonts, icons, desktop files."),
            ("tests", "Files to run tests & the GIT-LFS test data."),
            ("scripts", "Bundled Python scripts for UI layout, key-map & some operators."),
            ("tools", "Utilities to help with Blender development."),
        ],
    ),

)


# -----------------------------------------------------------------------------
# Generic Utilities

def _function_id() -> str:
    '''
    Create a string naming the function n frames up on the stack.
    '''
    # pylint: disable-next=protected-access
    co = sys._getframe(1).f_code
    return '{:s}:{:d}:'.format(co.co_name, co.co_firstlineno)


def object_data_materials_setup_default(object_data):
    object_data.materials.append(MATERIAL_FROM_COLOR["black"])
    object_data.materials.append(MATERIAL_FROM_COLOR["grey"])


# -----------------------------------------------------------------------------
# Internal Utility Classes

@dataclass
class Box2D:
    min_x: float
    max_x: float
    min_y: float
    max_y: float

    def size(self):
        return self.max_x - self.min_x, self.max_y - self.min_y

    def union(self, *rest):
        min_x, max_x, min_y, max_y = self.min_x, self.max_x, self.min_y, self.max_y
        for other in rest:
            min_x = min(min_x, other.min_x)
            max_x = max(max_x, other.max_x)
            min_y = min(min_y, other.min_y)
            max_y = max(max_y, other.max_y)
        return Box2D(min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)

    def expanded(self, value):
        result = self.copy()
        result.min_x -= value
        result.max_x += value
        result.min_y -= value
        result.max_y += value
        return result

    def expanded_x(self, value):
        result = self.copy()
        result.min_x -= value
        result.max_x += value
        return result

    def expanded_y(self, value):
        result = self.copy()
        result.min_y -= value
        result.max_y += value
        return result

    def copy(self, *, min_x=None, max_x=None, min_y=None, max_y=None):
        result = self.union()
        if min_x is not None:
            result.min_x = min_x
        if max_x is not None:
            result.max_x = max_x
        if min_y is not None:
            result.min_y = min_y
        if max_y is not None:
            result.max_y = max_y
        return result

    def isect_x(self, other):
        if other.max_x < self.min_x:
            return False
        if self.max_x < other.min_x:
            return False
        return True


@dataclass
class TextStyle:
    font: str
    size: float

    @property
    def line_height(self):
        return self.size * 0.01

    def apply(self, txt_data):
        settings = FONT_STYLE_SETTINGS[self.font]

        scale = 0.01 * settings["scale"]
        space_line = settings["space_line"]

        txt_data.size = self.size * scale
        txt_data.space_line = space_line
        txt_data.font = VFONT_FROM_STYLE[self.font]


text_style_title = TextStyle(font="default", size=20.0)
text_style_subheading = TextStyle(font="default", size=14.0)
text_style_body = TextStyle(font="default", size=8.0)
text_style_dir_title = TextStyle(font="mono", size=8.0)
text_style_dir_body = TextStyle(font="default", size=6.0)


# -----------------------------------------------------------------------------
# Generic Drawing Utilities

def box_2d_from_vectors(vectors):
    min_x = max_x = vectors[0][0]
    min_y = max_y = vectors[0][1]
    for i in range(1, len(vectors)):
        x, y = vectors[i][0:2]
        min_x = min(x, min_x)
        max_x = max(x, max_x)
        min_y = min(y, min_y)
        max_y = max(y, max_y)
    return Box2D(min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)


def box_2d_from_vectors_with_matrix(vectors, matrix):
    return box_2d_from_vectors([matrix @ Vector(v) for v in vectors])


def draw_text_centered(scene, *, location, text, style):
    name_gen = _function_id() + repr(text)
    txt_data = bpy.data.curves.new(name=name_gen, type='FONT')
    object_data_materials_setup_default(txt_data)

    # Text Object
    txt_data.body = text
    style.apply(txt_data)
    txt_data.align_x = 'CENTER'
    txt_ob = bpy.data.objects.new(name=name_gen, object_data=txt_data)
    txt_ob.location.xy = location

    scene.collection.objects.link(txt_ob)

    bpy.context.view_layer.update()
    return box_2d_from_vectors_with_matrix(txt_ob.bound_box, txt_ob.matrix_world.copy()), txt_ob


def draw_text_left(scene, *, location, text, style):
    name_gen = _function_id() + repr(text)
    txt_data = bpy.data.curves.new(name=name_gen, type='FONT')
    object_data_materials_setup_default(txt_data)

    # Text Object
    txt_data.body = text
    style.apply(txt_data)
    txt_data.align_x = 'LEFT'
    txt_ob = bpy.data.objects.new(name=name_gen, object_data=txt_data)
    txt_ob.location.xy = location

    scene.collection.objects.link(txt_ob)

    bpy.context.view_layer.update()
    return box_2d_from_vectors_with_matrix(txt_ob.bound_box, txt_ob.matrix_world.copy()), txt_ob


def draw_text_left_with_bounds(scene, *, box, text, style):
    name_gen = _function_id() + repr(text)
    txt_data = bpy.data.curves.new(name=name_gen, type='FONT')
    object_data_materials_setup_default(txt_data)

    # Text Object
    txt_data.body = text
    style.apply(txt_data)
    txt_data.align_x = 'LEFT'
    txt_box = txt_data.text_boxes[0]
    txt_box.x = box.min_x
    txt_box.y = box.max_y - style.line_height
    txt_box.width, txt_box.height = box.size()

    txt_ob = bpy.data.objects.new(name=name_gen, object_data=txt_data)

    scene.collection.objects.link(txt_ob)

    bpy.context.view_layer.update()
    return box_2d_from_vectors_with_matrix(txt_ob.bound_box, txt_ob.matrix_world.copy()), txt_ob


def draw_arrow(scene, *, location, size, line_width, arrow_data):
    name_gen = _function_id() + repr(location)

    import bmesh
    bm = bmesh.new()

    verts = [
        bm.verts.new(((co[0] * size) + location[0], (co[1] * size) + location[1], 0.0))
        for co in arrow_data
    ]

    f = bm.faces.new(verts)
    bm.normal_update()

    bmesh.ops.inset_individual(
        bm,
        faces=[f],
        use_even_offset=True,
        use_relative_offset=False,
        thickness=line_width,
    )

    bm.faces.remove(f)

    me = bpy.data.meshes.new(name=name_gen)
    object_data_materials_setup_default(me)
    bm.to_mesh(me)
    bm.free()

    line_ob = bpy.data.objects.new(name=name_gen, object_data=me)
    scene.collection.objects.link(line_ob)


# -----------------------------------------------------------------------------
# Setup

def setup_page():
    pass


# -----------------------------------------------------------------------------
# Drawing

def draw_mesh_dashed_line(bm, *, min_x, max_x, y, dash_on, dash_off, line_width, skip_box):
    x = min_x + (dash_off / 2)
    line_width_half = line_width / 2
    while x < max_x:
        beg_x = x
        end_x = x + dash_on
        if not skip_box.isect_x(Box2D(min_x=beg_x, max_x=end_x, min_y=0.0, max_y=0.0)):
            quad = [bm.verts.new() for i in range(4)]
            quad[0].co.xy = beg_x, y + line_width_half
            quad[1].co.xy = beg_x, y - line_width_half
            quad[2].co.xy = end_x, y - line_width_half
            quad[3].co.xy = end_x, y + line_width_half
            bm.faces.new(quad)
        x += dash_on + dash_off


def draw_mesh_box(bm, *, box, material_index=MATERIAL_INDEX_BLACK):
    quad = [bm.verts.new() for i in range(4)]
    quad[0].co.xy = box.min_x, box.min_y
    quad[1].co.xy = box.max_x, box.min_y
    quad[2].co.xy = box.max_x, box.max_y
    quad[3].co.xy = box.min_x, box.max_y
    f = bm.faces.new(quad)
    if material_index > 0:
        f.material_index = material_index


def draw_mesh_box_wire(bm, *, box, line_width):
    draw_mesh_box(bm, box=box.copy(max_x=box.min_x + line_width))
    draw_mesh_box(bm, box=box.copy(min_x=box.max_x - line_width))
    draw_mesh_box(bm, box=box.copy(max_y=box.min_y + line_width))
    draw_mesh_box(bm, box=box.copy(min_y=box.max_y - line_width))


def draw_mesh_box_drop_shadow(bm, *, box, size):
    # Right hand side.
    draw_mesh_box(
        bm,
        box=box.copy(
            min_x=box.max_x,
            max_x=box.max_x + size,
            min_y=box.min_y - size,
            max_y=box.max_y - size,
        ),
        material_index=MATERIAL_INDEX_GREY,
    )
    # Bottom.
    draw_mesh_box(
        bm,
        box=box.copy(
            min_x=box.min_x + size,
            max_x=box.max_x + size,
            min_y=box.min_y - size,
            max_y=box.min_y,
        ),
        material_index=MATERIAL_INDEX_GREY,
    )


# -----------------------------------------------------------------------------
# Drawing

def draw_centered_title(scene, step_y, text):
    box, ob = draw_text_centered(scene, location=(0.0, step_y), text=text, style=text_style_title)
    return box, ob


def sub_section_line(scene, box, step_y):
    name_gen = _function_id() + repr(box)

    import bmesh

    bm = bmesh.new()

    arrow_h = 0.075
    arrow_w = 0.1

    for sign in (-1.0, 1.0):
        # Left triangle.
        vs = [bm.verts.new() for i in range(3)]
        l = PAGE_WIDTH_HALF * sign
        vs[0].co.xy = l, step_y
        vs[1].co.xy = l - (0.1 * sign), step_y - (arrow_h / 2)
        vs[2].co.xy = l - (0.1 * sign), step_y + (arrow_h / 2)
        bm.faces.new(vs)

    # Horizontal line.
    draw_mesh_dashed_line(
        bm,
        min_x=-(PAGE_WIDTH_HALF - arrow_w),
        max_x=+(PAGE_WIDTH_HALF - arrow_w),
        y=step_y,
        dash_on=0.0375,
        dash_off=0.0375,
        line_width=0.015,
        skip_box=box.expanded_x(0.0375),
    )

    me = bpy.data.meshes.new(name=name_gen)
    object_data_materials_setup_default(me)
    bm.to_mesh(me)
    bm.free()

    line_ob = bpy.data.objects.new(name=name_gen, object_data=me)
    scene.collection.objects.link(line_ob)


def draw_directories(scene, box_body, source_code_dirs):
    name_gen = _function_id() + repr(source_code_dirs)

    import bmesh

    box_result = box_body.copy()

    box_line_width = 0.01
    box_size = 0.75, 0.3
    box_gap = 0.075, 0.075
    box_inner_margin = 0.04
    bm = bmesh.new()
    box_dropshadow_size = 0.025

    me = bpy.data.meshes.new(name=name_gen)

    step_x = box_body.min_x
    step_y = box_body.min_y

    # Adjust the box height based on the bounds of the body-text
    # (requires postponing drawing the boxes so the size is known).
    USE_FLEXIBLE_SIZE_X = True
    # Use double-width boxes when the directory name doesn't fit.
    USE_FLEXIBLE_SIZE_Y = True

    boxes_pending = []

    def draw_directory_boxes_pending():
        if not boxes_pending:
            return None
        y = boxes_pending[0][1].min_y
        for _, box_text in boxes_pending:
            y = min(y, box_text.min_y)
        y -= box_inner_margin

        for box_draw, _ in boxes_pending:
            box = box_draw.copy(min_y=y)
            draw_mesh_box_wire(bm, box=box, line_width=box_line_width)
            draw_mesh_box_drop_shadow(bm, box=box, size=box_dropshadow_size)

        bounds = boxes_pending[0][0].union(*[box_draw for box_draw, _ in boxes_pending])
        bounds.min_y = y

        boxes_pending.clear()
        return bounds

    for dir_name, description in source_code_dirs:
        box_size_flex = box_size

        # Draw text.
        box_dir, box_dir_ob = draw_text_left(
            scene,
            # First get the bounds.
            location=(0.0, 0.0),
            text=dir_name,
            style=text_style_dir_title,
        )
        if USE_FLEXIBLE_SIZE_X:
            if box_size_flex[0] < (box_dir.size()[0] + box_inner_margin * 2.0):
                box_size_flex = box_size_flex[0] + box_gap[0] + box_size[0], box_size[1]

        if step_x + box_size_flex[0] + box_gap[0] > box_body.max_x:
            step_y_size = box_size_flex[1]
            if USE_FLEXIBLE_SIZE_Y:
                box_pending = draw_directory_boxes_pending()
                if box_pending is not None:
                    step_y_size = box_pending.size()[1]
            step_y -= step_y_size + box_gap[1]
            step_x = box_body.min_x

        box_draw = Box2D(min_x=step_x, max_x=step_x + box_size_flex[0], min_y=step_y - box_size_flex[1], max_y=step_y)
        if not USE_FLEXIBLE_SIZE_Y:
            draw_mesh_box_wire(bm, box=box_draw, line_width=box_line_width)
            draw_mesh_box_drop_shadow(bm, box=box_draw, size=box_dropshadow_size)

        next_y = step_y - ((text_style_dir_title.line_height * 0.75) + box_inner_margin)
        next_y_body_begin = step_y - (text_style_dir_title.line_height + box_inner_margin)

        box_dir_ob.location.xy = (step_x + box_inner_margin, next_y)

        if description:
            box_text_bounds, _ = draw_text_left_with_bounds(
                scene,
                box=Box2D(
                    min_x=box_draw.min_x + box_inner_margin,
                    max_x=box_draw.max_x - box_inner_margin,
                    min_y=box_draw.min_y,
                    max_y=next_y_body_begin,
                ),
                text=description,
                style=text_style_dir_body,
            )
        if USE_FLEXIBLE_SIZE_Y:
            if description:
                boxes_pending.append((box_draw, box_text_bounds))
            else:
                boxes_pending.append((box_draw, box_draw.copy(min_y=next_y_body_begin)))

        step_x += box_size_flex[0] + box_gap[0]

    step_y_size = box_size[1]
    if USE_FLEXIBLE_SIZE_Y:
        box_pending = draw_directory_boxes_pending()
        if box_pending is not None:
            step_y_size = box_pending.size()[1]

    box_result.min_y = step_y - (step_y_size + box_gap[1])

    object_data_materials_setup_default(me)
    bm.to_mesh(me)
    bm.free()

    line_ob = bpy.data.objects.new(name=name_gen, object_data=me)
    scene.collection.objects.link(line_ob)

    return box_result


def draw_legend(scene, *, step_y, legend_data, text):

    unit = PAGE_WIDTH / 16.0

    x_left = -PAGE_WIDTH_HALF + (unit * 3)

    draw_arrow(
        scene,
        location=(x_left, step_y - 0.25),
        size=0.0575,
        line_width=0.0075,
        arrow_data=legend_data,
    )

    # Draw text.
    box_dir, _box_dir_ob = draw_text_left(
        scene,
        # First get the bounds.
        location=(x_left + 0.25, step_y - 0.175),
        text=text,
        style=text_style_subheading,
    )

    box_body = Box2D(
        min_x=-PAGE_WIDTH_HALF + unit * 3,
        max_x=box_dir.max_x,
        min_y=step_y - 0.25,
        max_y=step_y,
    )

    return box_body


def draw_sub_section(scene, *, step_y, section_data,):

    box_sub_heading, _ = draw_text_centered(
        scene,
        # The 0.025 tweak is needed to be vertically centered.
        # This depends on the exact font used.
        location=(0.0, step_y + 0.025),
        text=section_data.heading,
        style=text_style_subheading,
    )

    # Add dashes and arrows around.
    sub_section_line(scene, box_sub_heading, step_y + 0.05)

    step_y -= 0.2

    box_dir, _ = draw_text_left(
        scene,
        location=(-PAGE_WIDTH_HALF, step_y),
        text=section_data.source_code_base,
        style=text_style_body,
    )

    step_y += 0.1

    # Box with no height to draw mini boxes into.
    unit = PAGE_WIDTH / 16.0
    box_body = Box2D(
        min_x=-PAGE_WIDTH_HALF + unit * 3,
        max_x=PAGE_WIDTH_HALF - unit,
        min_y=step_y,
        max_y=step_y,
    )

    box_dirs = draw_directories(scene, box_body, section_data.source_code_dirs)

    # Draw an arrow on the right hand side.
    if section_data.source_code_call_sibling_modules is not None:
        draw_arrow(
            scene,
            location=(PAGE_WIDTH_HALF - 0.175, step_y - 0.25),
            size=0.0575,
            line_width=0.0075,
            arrow_data=(
                ARROW_DOWN_AND_SIDEWAYS if section_data.source_code_call_sibling_modules else
                ARROW_DOWN
            ),
        )

    return box_sub_heading.union(box_dir, box_body, box_dirs)


# -----------------------------------------------------------------------------
# Render Output

def render_output(scene, bounds, filepath):
    name_gen = _function_id()

    scene.render.filepath = filepath

    world = bpy.data.worlds.new(name_gen)
    world.node_tree.nodes.clear()
    output = world.node_tree.nodes.new("ShaderNodeOutputWorld")
    background = world.node_tree.nodes.new("ShaderNodeBackground")
    world.node_tree.links.new(output.outputs["Surface"], background.outputs["Surface"])
    background.inputs["Color"].default_value = 1.0, 1.0, 1.0, 1.0
    scene.world = world

    # Some space around the edges.
    bounds_size = bounds.expanded(0.15).size()

    camera_data = bpy.data.cameras.new(name_gen)
    camera_data.type = 'ORTHO'
    camera_data.ortho_scale = max(bounds_size)

    camera = bpy.data.objects.new(name_gen, camera_data)
    camera.rotation_euler = Euler((0.0, 0.0, 0.0), 'XYZ')

    scene.camera = camera
    camera.location = (0.0, (bounds.min_y + bounds.max_y) / 2.0, 10.0)
    scene.collection.objects.link(camera)

    render = scene.render
    render.image_settings.media_type = 'IMAGE'
    render.image_settings.file_format = 'JPEG'
    render.image_settings.color_depth = '8'
    render.image_settings.color_mode = 'RGB'
    render.image_settings.quality = 75
    render.use_file_extension = False
    render.resolution_x = RESOLUTION_X
    render.resolution_y = int(RESOLUTION_X * (bounds_size[1] / bounds_size[0]))
    render.resolution_percentage = int(100 * RESOLUTION_SCALE)

    bpy.context.view_layer.update()

    if OUTPUT_BLEND_FILE:
        bpy.ops.wm.save_mainfile('EXEC_DEFAULT', filepath=os.path.splitext(filepath)[0] + ".blend")

    if OUTPUT_IMAGE_FILE:
        print("Rendering:", filepath)
        bpy.ops.render.render(write_still=True)

# -----------------------------------------------------------------------------
# Validate Sections


def validate_sections():
    sections_shared = {section.source_code_base: [] for section in SECTIONS}

    for section in SECTIONS:
        if section.heading == "Operating system":
            continue
        sections_shared[section.source_code_base].append(section)

    # Consider the parent directories "documented".
    # Because their children are included.
    dirs_parent_documented = set()
    for dirpath in sections_shared.keys():
        if dirpath == "":
            continue
        if not dirpath.endswith("/"):
            raise Exception("Directories must end with a \"/\", found {:s}".format(dirpath))
        dirpath_split = dirpath.strip("/").split("/")
        for i in range(1, len(dirpath_split) + 1):
            dirpath = "/".join(dirpath_split[0:i])
            dirs_parent_documented.add(dirpath)

    this_platform = ""
    for e in os.scandir(os.path.join(ROOT_DIR, "lib")):
        dirpath_full = os.path.join(ROOT_DIR, "lib", e.name)
        if os.path.isdir(os.path.join(dirpath_full, "jpeg")):
            this_platform = e.name
            break

    dirs_fs_ignore = {
        ".git",
        ".github",
        ".gitea",
        ".well-known",
    }

    for source_code_base, sections in sections_shared.items():

        if "{platform}" in source_code_base:
            source_code_base = source_code_base.replace("{platform}", this_platform)

        dirs_fs = {
            e.name for e in os.scandir(os.path.join(ROOT_DIR, source_code_base))
            if e.is_dir()
        }

        for section in sections:
            for dir_docs, _ in section.source_code_dirs:
                if dir_docs in dirs_fs:
                    dirs_fs.remove(dir_docs)
                    continue

                print("Directory no longer exists:", os.path.join(source_code_base, dir_docs))

        for dir_fs in sorted(dirs_fs):
            if dir_fs in dirs_fs_ignore:
                continue
            if dir_fs in dirs_parent_documented:
                continue
            print("Directory has no docs:", os.path.join(source_code_base, dir_fs))


def argparse_create():
    import argparse

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--only-validate",
        dest="only_validate",
        action='store_true',
        help="Validate the directory listing and exit.",
    )
    parser.add_argument(
        "--output",
        dest="output",
        default="code_layout.jpg",
        type=str,
        help="The output path to write the JPEG to.",
    )

    return parser


# -----------------------------------------------------------------------------
# Main

def main():
    args = argparse_create().parse_args((sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else []))

    # Always validate as it's cheap and notifies of incomplete docs.
    validate_sections()
    if args.only_validate:
        print("Validation complete, exiting!")
        return

    bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)

    # Setup materials.
    material = bpy.data.materials.new("Flat Black")
    MATERIAL_FROM_COLOR["black"] = material
    del material
    material = bpy.data.materials.new("Flat Grey")
    nodes = material.node_tree.nodes
    bsdf = nodes.new("ShaderNodeBsdfPrincipled")
    output = nodes.new("ShaderNodeOutputMaterial")
    material.node_tree.links.new(bsdf.outputs["BSDF"], output.inputs["Surface"])
    bsdf.inputs['Base Color'].default_value = (0.4, 0.4, 0.4, 1.0)
    MATERIAL_FROM_COLOR["grey"] = material
    del material

    # Setup fonts.
    VFONT_FROM_STYLE["default"] = bpy.data.fonts.load(FONT_FILE_DEFAULT)
    VFONT_FROM_STYLE["mono"] = bpy.data.fonts.load(FONT_FILE_MONO)

    scene = bpy.context.scene
    scene.render.engine = 'BLENDER_EEVEE'

    # Without this, the whites are gray.
    scene.view_settings.view_transform = "Standard"

    sub_section_gap_y = 0.1

    setup_page()
    step_y = 0.0
    box, _ = draw_centered_title(scene, step_y, "Blender Code Layout")
    bounds_max_y = box.max_y

    step_y -= box.max_y - box.min_y

    # Draw the legends for each kind of arrow.
    box = draw_legend(
        scene,
        step_y=step_y,
        legend_data=ARROW_DOWN,
        text="Modules only call lower level code.",
    )
    step_y = box.min_y - 0.175
    box = draw_legend(
        scene,
        step_y=step_y,
        legend_data=ARROW_DOWN_AND_SIDEWAYS,
        text="Modules call each other and lower level code.",
    )
    step_y = box.min_y - (0.175 + 0.15)

    # Draw the sections.
    for section_data in SECTIONS:
        box = draw_sub_section(
            scene,
            step_y=step_y,
            section_data=section_data,
        )
        step_y = box.min_y - sub_section_gap_y

    # Setup Y resolutions.
    bounds = Box2D(min_x=-PAGE_WIDTH_HALF, max_x=PAGE_WIDTH_HALF, min_y=box.min_y, max_y=bounds_max_y)

    print(args.output)
    render_output(scene, bounds, args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This script takes 2-3 args: [--browse] <Doxyfile> <sourcefile>

Where Doxyfile is a path relative to source root,
and the sourcefile as an absolute path.

--browse will open the resulting docs in a web browser.
"""
import sys
import os
import subprocess
import tempfile


def find_gitroot(filepath_reference):
    path = filepath_reference
    path_prev = ""
    while not os.path.exists(os.path.join(path, ".git")) and path != path_prev:
        path_prev = path
        path = os.path.dirname(path)
    return path


doxyfile, sourcefile = sys.argv[-2:]

doxyfile = os.path.join(find_gitroot(sourcefile), doxyfile)
os.chdir(os.path.dirname(doxyfile))

tempfile = tempfile.NamedTemporaryFile(mode='w+b')
doxyfile_tmp = tempfile.name
tempfile.write(open(doxyfile, "r+b").read())
tempfile.write(b'\n\n')
tempfile.write(b'INPUT=' + os.fsencode(sourcefile) + b'\n')
tempfile.flush()

subprocess.call(("doxygen", doxyfile_tmp))
del tempfile

# Maybe handy, but also annoying?
if "--browse" in sys.argv:
    import webbrowser
    webbrowser.open("html/files.html")



# SPDX-FileCopyrightText: 2018-2023 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

"""
Compare renders or screenshots against reference versions and generate
a HTML report showing the differences, for regression testing.
"""

import glob
import os
import sys
import pathlib
import shutil
import subprocess
import time
import multiprocessing

from . import global_report
from .colored_print import (print_message, use_message_colors)


def blend_list(dirpath, blocklist):
    import re

    for root, dirs, files in os.walk(dirpath):
        for filename in files:
            if not filename.lower().endswith(".blend"):
                continue

            skip = False
            for blocklist_entry in blocklist:
                if re.match(blocklist_entry, filename):
                    skip = True
                    break

            if not skip:
                filepath = os.path.join(root, filename)
                yield filepath


def test_get_name(filepath):
    filename = os.path.basename(filepath)
    return os.path.splitext(filename)[0]


def test_get_images(output_dir, filepath, testname, reference_dir, reference_override_dir):
    dirpath = os.path.dirname(filepath)

    old_dirpath = os.path.join(dirpath, reference_dir)
    old_img = os.path.join(old_dirpath, testname + ".png")
    if reference_override_dir:
        override_dirpath = os.path.join(dirpath, reference_override_dir)
        override_img = os.path.join(override_dirpath, testname + ".png")
        if os.path.exists(override_img):
            old_dirpath = override_dirpath
            old_img = override_img

    ref_dirpath = os.path.join(output_dir, os.path.basename(dirpath), "ref")
    ref_img = os.path.join(ref_dirpath, testname + ".png")
    os.makedirs(ref_dirpath, exist_ok=True)
    if os.path.exists(old_img):
        shutil.copy(old_img, ref_img)

    new_dirpath = os.path.join(output_dir, os.path.basename(dirpath))
    os.makedirs(new_dirpath, exist_ok=True)
    new_img = os.path.join(new_dirpath, testname + ".png")

    diff_dirpath = os.path.join(output_dir, os.path.basename(dirpath), "diff")
    os.makedirs(diff_dirpath, exist_ok=True)
    diff_color_img = os.path.join(diff_dirpath, testname + ".diff_color.png")
    diff_alpha_img = os.path.join(diff_dirpath, testname + ".diff_alpha.png")

    return old_img, ref_img, new_img, diff_color_img, diff_alpha_img


class TestResult:
    def __init__(self, report, filepath, name):
        self.filepath = filepath
        self.name = name
        self.error = None
        self.tmp_out_img_base = os.path.join(report.output_dir, "tmp_" + name)
        self.tmp_out_img = self.tmp_out_img_base + '0001.png'
        self.old_img, self.ref_img, self.new_img, self.diff_color_img, self.diff_alpha_img = test_get_images(
            report.output_dir, filepath, name, report.reference_dir, report.reference_override_dir)


def diff_output(test, oiiotool, fail_threshold, fail_percent, verbose, update):
    # Create reference render directory.
    old_dirpath = os.path.dirname(test.old_img)
    os.makedirs(old_dirpath, exist_ok=True)

    # Copy temporary to new image.
    if os.path.exists(test.new_img):
        os.remove(test.new_img)
    if os.path.exists(test.tmp_out_img):
        shutil.copy(test.tmp_out_img, test.new_img)

    if os.path.exists(test.ref_img):
        # Diff images test with threshold.
        command = (
            oiiotool,
            test.ref_img,
            test.tmp_out_img,
            "--fail", str(fail_threshold),
            "--failpercent", str(fail_percent),
            "--diff",
        )
        try:
            subprocess.check_output(command)
            failed = False
        except subprocess.CalledProcessError as e:
            if verbose:
                print_message(e.output.decode("utf-8", 'ignore'))
            failed = e.returncode != 0
    else:
        if not update:
            test.error = "VERIFY"
            return test

        failed = True

    if failed and update:
        # Update reference image if requested.
        shutil.copy(test.new_img, test.ref_img)
        shutil.copy(test.new_img, test.old_img)
        failed = False

    # Generate color diff image.
    command = (
        oiiotool,
        test.ref_img,
        "--ch", "R,G,B",
        test.tmp_out_img,
        "--ch", "R,G,B",
        "--sub",
        "--abs",
        "--mulc", "16",
        "-o", test.diff_color_img,
    )
    try:
        subprocess.check_output(command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if verbose:
            print_message(e.output.decode("utf-8", 'ignore'))

    # Generate alpha diff image.
    command = (
        oiiotool,
        test.ref_img,
        "--ch", "A",
        test.tmp_out_img,
        "--ch", "A",
        "--sub",
        "--abs",
        "--mulc", "16",
        "-o", test.diff_alpha_img,
    )
    try:
        subprocess.check_output(command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if verbose:
            msg = e.output.decode("utf-8", 'ignore')
            for line in msg.splitlines():
                # Ignore warnings for images without alpha channel.
                if "--ch: Unknown channel name" not in line:
                    print_message(line)

    if failed:
        test.error = "VERIFY"
    else:
        test.error = None

    return test


def get_gpu_device_vendor(blender):
    command = [
        blender,
        "--background",
        "--factory-startup",
        "--python",
        str(pathlib.Path(__file__).parent / "gpu_info.py")
    ]
    try:
        completed_process = subprocess.run(command, stdout=subprocess.PIPE, universal_newlines=True)
        for line in completed_process.stdout.splitlines():
            if line.startswith("GPU_DEVICE_TYPE:"):
                vendor = line.split(':')[1].upper()
                return vendor
    except Exception:
        return None
    return None


class Report:
    __slots__ = (
        'title',
        'engine_name',
        'output_dir',
        'global_dir',
        'reference_dir',
        'reference_override_dir',
        'oiiotool',
        'pixelated',
        'fail_threshold',
        'fail_percent',
        'verbose',
        'update',
        'failed_tests',
        'passed_tests',
        'compare_tests',
        'compare_engine',
        'blocklist',
    )

    def __init__(self, title, output_dir, oiiotool, variation=None, blocklist=[]):
        self.title = title

        # Normalize the path to avoid output_dir and global_dir being the same when a directory
        # ends with a trailing slash.
        self.output_dir = os.path.normpath(output_dir)
        self.global_dir = os.path.dirname(self.output_dir)

        self.reference_dir = 'reference_renders'
        self.reference_override_dir = None
        self.oiiotool = oiiotool
        self.compare_engine = None
        self.fail_threshold = 0.016
        self.fail_percent = 1
        self.engine_name = self.title.lower().replace(" ", "_")
        self.blocklist = [] if os.getenv('BLENDER_TEST_IGNORE_BLOCKLIST') is not None else blocklist

        if variation:
            self.title = self._engine_title(title, variation)
            self.output_dir = self._engine_path(self.output_dir, variation.lower())

        self.pixelated = False
        self.verbose = os.environ.get("BLENDER_VERBOSE") is not None
        self.update = os.getenv('BLENDER_TEST_UPDATE') is not None

        if os.environ.get("BLENDER_TEST_COLOR") is not None:
            use_message_colors()

        self.failed_tests = ""
        self.passed_tests = ""
        self.compare_tests = ""

        os.makedirs(output_dir, exist_ok=True)

    def set_pixelated(self, pixelated):
        self.pixelated = pixelated

    def set_fail_threshold(self, threshold):
        self.fail_threshold = threshold

    def set_fail_percent(self, percent):
        self.fail_percent = percent

    def set_reference_dir(self, reference_dir):
        self.reference_dir = reference_dir

    def set_reference_override_dir(self, reference_override_dir):
        self.reference_override_dir = reference_override_dir

    def set_compare_engine(self, other_engine, other_variation=None):
        self.compare_engine = (other_engine, other_variation)

    def set_engine_name(self, engine_name):
        self.engine_name = engine_name

    def run(self, dirpath, blender, arguments_cb, batch=False, fail_silently=False):
        # Run tests and output report.
        dirname = os.path.basename(dirpath)
        ok = self._run_all_tests(dirname, dirpath, blender, arguments_cb, batch, fail_silently)
        self._write_data(dirname)
        self._write_html()
        if self.compare_engine:
            self._write_html(comparison=True)
        return ok

    def _write_data(self, dirname):
        # Write intermediate data for single test.
        outdir = os.path.join(self.output_dir, dirname)
        os.makedirs(outdir, exist_ok=True)

        filepath = os.path.join(outdir, "failed.data")
        pathlib.Path(filepath).write_text(self.failed_tests)

        filepath = os.path.join(outdir, "passed.data")
        pathlib.Path(filepath).write_text(self.passed_tests)

        if self.compare_engine:
            filepath = os.path.join(outdir, "compare.data")
            pathlib.Path(filepath).write_text(self.compare_tests)

    def _navigation_item(self, title, href, active):
        if active:
            return """<li class="breadcrumb-item active" aria-current="page">%s</li>""" % title
        else:
            return """<li class="breadcrumb-item"><a href="%s">%s</a></li>""" % (href, title)

    def _engine_title(self, engine, variation):
        if variation:
            return engine.title() + ' ' + variation
        else:
            return engine.title()

    def _engine_path(self, path, variation):
        if variation:
            variation = variation.replace(' ', '_')
            return os.path.join(path, variation.lower())
        else:
            return path

    def _navigation_html(self, comparison):
        html = """<nav aria-label="breadcrumb"><ol class="breadcrumb">"""
        base_path = os.path.relpath(self.global_dir, self.output_dir)
        global_report_path = os.path.join(base_path, "report.html")
        html += self._navigation_item("Test Reports", global_report_path, False)
        html += self._navigation_item(self.title, "report.html", not comparison)
        if self.compare_engine:
            compare_title = "Compare with %s" % self._engine_title(*self.compare_engine)
            html += self._navigation_item(compare_title, "compare.html", comparison)
        html += """</ol></nav>"""

        return html

    def _write_html(self, comparison=False):
        # Gather intermediate data for all tests.
        if comparison:
            failed_data = []
            passed_data = sorted(glob.glob(os.path.join(self.output_dir, "*/compare.data")))
        else:
            failed_data = sorted(glob.glob(os.path.join(self.output_dir, "*/failed.data")))
            passed_data = sorted(glob.glob(os.path.join(self.output_dir, "*/passed.data")))

        failed_tests = ""
        passed_tests = ""

        for filename in failed_data:
            filepath = os.path.join(self.output_dir, filename)
            failed_tests += pathlib.Path(filepath).read_text()
        for filename in passed_data:
            filepath = os.path.join(self.output_dir, filename)
            passed_tests += pathlib.Path(filepath).read_text()

        tests_html = failed_tests + passed_tests

        # Write html for all tests.
        if self.pixelated:
            image_rendering = 'pixelated'
        else:
            image_rendering = 'auto'

        # Navigation
        menu = self._navigation_html(comparison)

        failed = len(failed_tests) > 0
        if failed:
            message = """<div class="alert alert-danger" role="alert">"""
            message += """<p>Run this command to regenerate reference (ground truth) images:</p>"""
            message += """<p><tt>BLENDER_TEST_UPDATE=1 ctest -R %s</tt></p>""" % self.engine_name
            message += """<p>This then happens for new and failing tests; reference images of """ \
                       """passing test cases will not be updated. Be sure to commit the new reference """ \
                       """images under the tests/files folder afterwards.</p>"""
            message += """</div>"""
        else:
            message = ""

        if comparison:
            title = self.title + " Test Compare"
            engine_self = self.title
            engine_other = self._engine_title(*self.compare_engine)
            columns_html = "<tr><th>Name</th><th>%s</th><th>%s</th>" % (engine_self, engine_other)
        else:
            title = self.title + " Test Report"
            columns_html = "<tr><th>Name</th><th>New</th><th>Reference</th><th>Diff Color</th><th>Diff Alpha</th>"

        html = f"""
<html>
<head>
    <title>{title}</title>
    <style>
        div.page_container {{
          text-align: center;
        }}
        div.page_container div {{
          text-align: left;
        }}
        div.page_content {{
          display: inline-block;
        }}
        img {{ image-rendering: {image_rendering}; width: 256px; background-color: #000; }}
        img.render {{
            background-color: #fff;
            background-image:
              -moz-linear-gradient(45deg, #eee 25%, transparent 25%),
              -moz-linear-gradient(-45deg, #eee 25%, transparent 25%),
              -moz-linear-gradient(45deg, transparent 75%, #eee 75%),
              -moz-linear-gradient(-45deg, transparent 75%, #eee 75%);
            background-image:
              -webkit-gradient(linear, 0 100%, 100% 0, color-stop(.25, #eee), color-stop(.25, transparent)),
              -webkit-gradient(linear, 0 0, 100% 100%, color-stop(.25, #eee), color-stop(.25, transparent)),
              -webkit-gradient(linear, 0 100%, 100% 0, color-stop(.75, transparent), color-stop(.75, #eee)),
              -webkit-gradient(linear, 0 0, 100% 100%, color-stop(.75, transparent), color-stop(.75, #eee));

            -moz-background-size:50px 50px;
            background-size:50px 50px;
            -webkit-background-size:50px 51px; /* Override value for silly webkit. */

            background-position:0 0, 25px 0, 25px -25px, 0px 25px;
        }}
        table td:first-child {{ width: 256px; }}
        p {{ margin-bottom: 0.5rem; }}
    </style>
    <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">
</head>
<body>
    <div class="page_container"><div class="page_content">
        <br/>
        <h1>{title}</h1>
        {menu}
        {message}
        <table class="table table-striped">
            <thead class="thead-dark">
                {columns_html}
            </thead>
            {tests_html}
        </table>
        <br/>
    </div></div>
</body>
</html>
            """

        filename = "report.html" if not comparison else "compare.html"
        filepath = os.path.join(self.output_dir, filename)
        pathlib.Path(filepath).write_text(html)

        print_message("Report saved to: " + pathlib.Path(filepath).as_uri())

        # Update global report
        if not comparison:
            global_failed = failed if not comparison else None
            global_report.add(self.global_dir, "Render", self.title, filepath, global_failed)

    def _relative_url(self, filepath):
        relpath = os.path.relpath(filepath, self.output_dir)
        return pathlib.Path(relpath).as_posix()

    def _write_test_html(self, test_category, test_result):
        name = test_result.name.replace('_', ' ')

        status = test_result.error if test_result.error else ""
        tr_style = """ class="table-danger" """ if test_result.error else ""

        new_url = self._relative_url(test_result.new_img)
        ref_url = self._relative_url(test_result.ref_img)
        diff_color_url = self._relative_url(test_result.diff_color_img)
        diff_alpha_url = self._relative_url(test_result.diff_alpha_img)

        test_html = f"""
            <tr{tr_style}>
                <td><b>{name}</b><br/>{test_category}<br/>{status}</td>
                <td><img src="{new_url}" onmouseover="this.src='{ref_url}';" onmouseout="this.src='{new_url}';" class="render"></td>
                <td><img src="{ref_url}" onmouseover="this.src='{new_url}';" onmouseout="this.src='{ref_url}';" class="render"></td>
                <td><img src="{diff_color_url}"></td>
                <td><img src="{diff_alpha_url}"></td>
            </tr>"""

        if test_result.error:
            self.failed_tests += test_html
        else:
            self.passed_tests += test_html

        if self.compare_engine:
            base_path = os.path.relpath(self.global_dir, self.output_dir)
            ref_url = os.path.join(base_path, self._engine_path(*self.compare_engine), new_url)

            test_html = """
                <tr{tr_style}>
                    <td><b>{name}</b><br/>{testname}<br/>{status}</td>
                    <td><img src="{new_url}" onmouseover="this.src='{ref_url}';" onmouseout="this.src='{new_url}';" class="render"></td>
                    <td><img src="{ref_url}" onmouseover="this.src='{new_url}';" onmouseout="this.src='{ref_url}';" class="render"></td>
                </tr>""" . format(tr_style=tr_style,
                                  name=name,
                                  testname=test_result.name,
                                  status=status,
                                  new_url=new_url,
                                  ref_url=ref_url)

            self.compare_tests += test_html

    def _get_render_arguments(self, arguments_cb, filepath, base_output_filepath):
        # Each render test can override this method to provide extra functionality.
        # See Cycles render tests for an example.
        # Do not delete.
        return arguments_cb(filepath, base_output_filepath)

    def _get_arguments_suffix(self):
        # Get command line arguments that need to be provided after all file-specific ones.
        # For example the Cycles render device argument needs to be added at the end of
        # the argument list, otherwise tests can't be batched together.
        #
        # Each render test is supposed to override this method.
        return []

    def _get_filepath_tests(self, filepath):
        list_filepath = filepath.replace('.blend', '_permutations.txt')
        if os.path.exists(list_filepath):
            with open(list_filepath, 'r') as file:
                return [TestResult(self, filepath, testname.rstrip('\n')) for testname in file]
        else:
            testname = test_get_name(filepath)
            return [TestResult(self, filepath, testname)]

    def _run_tests(self, filepaths, blender, arguments_cb, batch):
        # Run multiple tests in a single Blender process since startup can be
        # a significant factor. In case of crashes, re-run the remaining tests.
        verbose = os.environ.get("BLENDER_VERBOSE") is not None

        remaining_filepaths = filepaths[:]
        test_results = []
        arguments_suffix = self._get_arguments_suffix()

        while len(remaining_filepaths) > 0:
            command = [blender]
            running_tests = []

            # On Windows, there is a maximum length of 32,767 characters (including the terminating null character)
            # for process command line commands, see:
            # https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessa
            command_line_length = len(blender)
            for suffix in arguments_suffix:
                # Add 3 for taking into account spaces and quotation marks potentially added by Python.
                command_line_length += len(suffix) + 3

            # Construct output filepaths and command to run
            for filepath in remaining_filepaths:
                testname = test_get_name(filepath)

                base_output_filepath = os.path.join(self.output_dir, "tmp_" + testname)
                command_filepath = self._get_render_arguments(arguments_cb, filepath, base_output_filepath)

                # Check if we have surpassed the command line limit.
                for cmd in command_filepath:
                    command_line_length += len(cmd) + 3
                if sys.platform == 'win32' and command_line_length > 32766 and len(running_tests) > 0:
                    break

                print_message(testname, 'SUCCESS', 'RUN')
                running_tests.append(filepath)
                command.extend(command_filepath)

                output_filepath = base_output_filepath + '0001.png'
                if os.path.exists(output_filepath):
                    os.remove(output_filepath)

                # Only chain multiple commands for batch
                if not batch:
                    break

            command.extend(arguments_suffix)

            # Run process
            crash = False
            output = None
            try:
                completed_process = subprocess.run(command, stdout=subprocess.PIPE)
                if completed_process.returncode != 0:
                    crash = True
                output = completed_process.stdout
            except Exception:
                crash = True

            if verbose:
                def quote_expr_args(cmd):
                    quoted = []
                    quote_next = False
                    for arg in cmd:
                        if quote_next:
                            quoted.append('"{}"'.format(arg))  # wrap the expression in quotes
                            quote_next = False
                        else:
                            quoted.append(arg)
                            if arg == "--python-expr":
                                quote_next = True
                    return quoted
                print(' '.join(quote_expr_args(command)))

            if (verbose or crash) and output:
                print(output.decode("utf-8", 'ignore'))

            tests_to_check = []

            # Detect missing filepaths and consider those errors
            for filepath in running_tests:
                remaining_filepaths.pop(0)
                file_crashed = False
                for test in self._get_filepath_tests(filepath):
                    self.postprocess_test(blender, test)
                    if not os.path.exists(test.tmp_out_img) or os.path.getsize(test.tmp_out_img) == 0:
                        if crash:
                            # In case of crash, stop after missing files and re-render remaining
                            test.error = "CRASH"
                            test_results.append(test)
                            file_crashed = True
                            break
                        else:
                            test.error = "NO OUTPUT"
                            test_results.append(test)
                    else:
                        tests_to_check.append(test)
                if file_crashed:
                    break

            pool = multiprocessing.Pool(multiprocessing.cpu_count())
            test_results.extend(pool.starmap(diff_output,
                                             [(test, self.oiiotool, self.fail_threshold, self.fail_percent, self.verbose, self.update)
                                              for test in tests_to_check]))
            pool.close()

        for test in test_results:
            if test.error == "CRASH":
                print_message("Crash running Blender")
                print_message(test.name, 'FAILURE', 'FAILED')
            elif test.error == "NO OUTPUT":
                print_message("No render result file found")
                print_message(test.tmp_out_img, 'FAILURE', 'FAILED')
            elif test.error == "VERIFY":
                print_message("Render result is different from reference image")
                print_message(test.name, 'FAILURE', 'FAILED')
            else:
                print_message(test.name, 'SUCCESS', 'OK')

            if os.path.exists(test.tmp_out_img):
                os.remove(test.tmp_out_img)

        return test_results

    def postprocess_test(self, blender, test):
        """
        Post-process test result after the Blender has run.
        For example, this function is where conversion from video to a still image suitable for image diffing.
        """

        pass

    def _run_all_tests(self, dirname, dirpath, blender, arguments_cb, batch, fail_silently):
        passed_tests = []
        failed_tests = []
        silently_failed_tests = []
        all_files = list(blend_list(dirpath, self.blocklist))
        all_files.sort()
        if not list(blend_list(dirpath, [])):
            print_message("No .blend files found in '{}'!".format(dirpath), 'FAILURE', 'FAILED')
            return False

        print_message("Running {} tests from 1 test case." .
                      format(len(all_files)),
                      'SUCCESS', "==========")
        time_start = time.time()
        test_results = self._run_tests(all_files, blender, arguments_cb, batch)
        for test in test_results:
            if test.error:
                if test.error == "NO_ENGINE":
                    return False
                elif test.error == "NO_START":
                    return False

                if fail_silently and test.error != 'CRASH':
                    silently_failed_tests.append(test.name)
                else:
                    failed_tests.append(test.name)
            else:
                passed_tests.append(test.name)
            self._write_test_html(dirname, test)
        time_end = time.time()
        elapsed_ms = int((time_end - time_start) * 1000)
        print_message("")
        print_message("{} tests from 1 test case ran. ({} ms total)" .
                      format(len(all_files), elapsed_ms),
                      'SUCCESS', "==========")
        print_message("{} tests." .
                      format(len(passed_tests)),
                      'SUCCESS', 'PASSED')
        all_failed_tests = silently_failed_tests + failed_tests
        if all_failed_tests:
            print_message("{} tests, listed below:" .
                          format(len(all_failed_tests)),
                          'FAILURE', 'FAILED')
            all_failed_tests.sort()
            for test in all_failed_tests:
                print_message("{}" . format(test), 'FAILURE', "FAILED")

        return not bool(failed_tests)
